"""Constants for scripts package."""

AUTO_GEN_POST = [
    "",
    "# ----------------------------------------------------------------------------------- #",
    "# **END OF AUTOGENERATED CONTENT** (DO NOT EDIT/REMOVE THIS COMMENT BLOCK AND DO NOT  #",
    "# EDIT ANYTHING ABOVE IT. IF A NEW IMPORT IS NEEDED, ADD IT TO THE IMPORTS IN THE     #",
    "# CORRESPONDING GENERATION SCRIPT THEN RE-RUN THE SCRIPT. LINES WRITTEN BELOW THIS    #",
    "# BLOCK WILL BE PRESERVED AS LONG AS THIS BLOCK REMAINS)                              #",
    "# ----------------------------------------------------------------------------------- #",
    "",
]
AUTO_GEN_PRE = [
    "",
    "# ----------------------------------------------------------------------------------- #",
    "# **BEGINNING OF AUTOGENERATED CONTENT** (TO ADD ADDITIONAL MANUAL CONTENT, LOOK FOR  #",
    '# THE "END OF AUTOGENERATED CONTENT" COMMENT BLOCK AND ADD YOUR CODE BELOW IT)        #',
    "# ----------------------------------------------------------------------------------- #",
    "",
]
GITHUB_PROJECT = "zwave-js/node-zwave-js"
BRANCH_NAME = "master"
