"""Constants for utility tests."""

from zwave_js_server.const.command_class.lock import (
    ATTR_CODE_SLOT,
    ATTR_IN_USE,
    ATTR_NAME,
    ATTR_USERCODE,
)

CODE_SLOTS = [
    {
        ATTR_CODE_SLOT: 1,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (1)",
        ATTR_USERCODE: "**********",
    },
    {
        ATTR_CODE_SLOT: 2,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (2)",
        ATTR_USERCODE: "**********",
    },
    {
        ATTR_CODE_SLOT: 3,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (3)",
        ATTR_USERCODE: "**********",
    },
    {
        ATTR_CODE_SLOT: 4,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (4)",
        ATTR_USERCODE: "7030\n\r",
    },
    {
        ATTR_CODE_SLOT: 5,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (5)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 6,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (6)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 7,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (7)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 8,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (8)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 9,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (9)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 10,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (10)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 11,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (11)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 12,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (12)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 13,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (13)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 14,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (14)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 15,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (15)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 16,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (16)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 17,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (17)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 18,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (18)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 19,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (19)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 20,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (20)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 21,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (21)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 22,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (22)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 23,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (23)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 24,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (24)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 25,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (25)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 26,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (26)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 27,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (27)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 28,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (28)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 29,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (29)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 30,
        ATTR_IN_USE: None,
        ATTR_NAME: "User Code (30)",
        ATTR_USERCODE: None,
    },
]
